@@ -5,11 +5,10 @@
 if [ -e ${prefix}/share/kea/scripts/admin-utils.sh ]; then
     . ${prefix}/share/kea/scripts/admin-utils.sh
 else
-    . /home/wlodek/dev/kea_rel/src/bin/admin/admin-utils.sh
+    . /home/wlodek/dev/kea_r/src/bin/admin/admin-utils.sh
 fi
 
-mysql_version "$@"
-VERSION=$_RESULT
+VERSION=`mysql_version "$@"`
 
 if [ "$VERSION" != "1.0" ]; then
     printf "This script upgrades 1.0 to 2.0. Reported version is $VERSION. Skipping upgrade.\n"
