@@ -13,11 +13,11 @@
 # PERFORMANCE OF THIS SOFTWARE.
 
 # Path to the temporary configuration file.
-CFG_FILE=/home/wlodek/dev/kea_rel/src/bin/dhcp6/tests/test_config.json
+CFG_FILE=/home/wlodek/dev/kea_r/src/bin/dhcp6/tests/test_config.json
 # Path to the Kea log file.
-LOG_FILE=/home/wlodek/dev/kea_rel/src/bin/dhcp6/tests/test.log
+LOG_FILE=/home/wlodek/dev/kea_r/src/bin/dhcp6/tests/test.log
 # Expected version
-EXPECTED_VERSION="0.9.1-beta"
+EXPECTED_VERSION="0.9.2-beta"
 # Kea configuration to be stored in the configuration file.
 CONFIG="{
     \"Dhcp6\":
@@ -95,10 +95,10 @@ CONFIG_INVALID="{
 
 # Set the location of the executable.
 bin="kea-dhcp6"
-bin_path=/home/wlodek/dev/kea_rel/src/bin/dhcp6
+bin_path=/home/wlodek/dev/kea_r/src/bin/dhcp6
 
 # Import common test library.
-. /home/wlodek/dev/kea_rel/src/lib/testutils/dhcp_test_lib.sh
+. /home/wlodek/dev/kea_r/src/lib/testutils/dhcp_test_lib.sh
 
 # This test verifies that DHCPv6 can be reconfigured with a SIGHUP signal.
 dynamic_reconfiguration_test() {
