@@ -12,19 +12,27 @@
 // OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 // PERFORMANCE OF THIS SOFTWARE.
 
+#include <config.h>
 #include <gtest/gtest.h>
+#include <dhcp/option6_status_code.h>
 #include <dhcp6/tests/dhcp6_test_utils.h>
 #include <dhcp6/json_config_parser.h>
-#include <config/ccsession.h>
+#include <dhcp/tests/pkt_captures.h>
+#include <util/pointer_util.h>
+#include <cc/command_interpreter.h>
+#include <stats/stats_mgr.h>
 #include <string.h>
 
 using namespace isc::data;
 using namespace isc::dhcp;
 using namespace isc::asiolink;
+using namespace isc::stats;
 
 namespace isc {
 namespace test {
 
+const char* NakedDhcpv6SrvTest::DUID_FILE = "server-id-test.txt";
+
 Dhcpv6SrvTest::Dhcpv6SrvTest()
 :srv_(0) {
     subnet_ = isc::dhcp::Subnet6Ptr
@@ -103,7 +111,7 @@ Dhcpv6SrvTest::checkIA_PD(const Pkt6Ptr& rsp, uint32_t expected_iaid,
 }
 
 // Checks if the lease sent to client is present in the database
-// and is valid when checked agasint the configured subnet
+// and is valid when checked against the configured subnet
 Lease6Ptr
 Dhcpv6SrvTest::checkLease(const DuidPtr& duid, const OptionPtr& ia_na,
                           boost::shared_ptr<Option6IAAddr> addr) {
@@ -133,7 +141,9 @@ Dhcpv6SrvTest::checkLease(const isc::dhcp::Lease6& lease) {
         return (Lease6Ptr());
     }
 
-    EXPECT_TRUE(lease_db->matches(lease));
+    EXPECT_TRUE(util::nullOrEqualValues(lease_db->hwaddr_, lease.hwaddr_));
+    EXPECT_TRUE(util::nullOrEqualValues(lease_db->duid_, lease.duid_));
+
     return (lease_db);
 }
 
@@ -585,6 +595,12 @@ Dhcpv6SrvTest::testReleaseBasic(Lease::Type type, const IOAddress& existing,
     Lease6Ptr l = LeaseMgrFactory::instance().getLease6(type, existing);
     ASSERT_TRUE(l);
 
+    // And prepopulate the stats counter
+    std::string name = StatsMgr::generateName("subnet", subnet_->getID(),
+                                              type == Lease::TYPE_NA ? "assigned-nas" :
+                                              "assigned-pds");
+    StatsMgr::instance().setValue(name, static_cast<int64_t>(1));
+
     // Let's create a RELEASE
     Pkt6Ptr rel = createMessage(DHCPV6_RELEASE, type, release_addr, prefix_len,
                                 iaid);
@@ -623,6 +639,11 @@ Dhcpv6SrvTest::testReleaseBasic(Lease::Type type, const IOAddress& existing,
     l = LeaseMgrFactory::instance().getLease6(type, *duid_, iaid,
                                               subnet_->getID());
     ASSERT_FALSE(l);
+
+    // We should have decremented the address counter
+    ObservationPtr stat = StatsMgr::instance().getObservation(name);
+    ASSERT_TRUE(stat);
+    EXPECT_EQ(0, stat->getInteger().first);
 }
 
 void
@@ -651,6 +672,12 @@ Dhcpv6SrvTest::testReleaseReject(Lease::Type type, const IOAddress& addr) {
     // GenerateClientId() also sets duid_
     OptionPtr clientid = generateClientId();
 
+    // Pretend we have allocated 1 lease
+    std::string name = StatsMgr::generateName("subnet", subnet_->getID(),
+                                              type == Lease::TYPE_NA ? "assigned-nas" :
+                                              "assigned-pds");
+    StatsMgr::instance().setValue(name, static_cast<int64_t>(1));
+
     // Check that the lease is NOT in the database
     Lease6Ptr l = LeaseMgrFactory::instance().getLease6(type, addr);
     ASSERT_FALSE(l);
@@ -680,6 +707,11 @@ Dhcpv6SrvTest::testReleaseReject(Lease::Type type, const IOAddress& addr) {
     l = LeaseMgrFactory::instance().getLease6(type, addr);
     ASSERT_FALSE(l);
 
+    // Verify we didn't decrement the stats counter
+    ObservationPtr stat = StatsMgr::instance().getObservation(name);
+    ASSERT_TRUE(stat);
+    EXPECT_EQ(1, stat->getInteger().first);
+
     // CASE 2: Lease is known and belongs to this client, but to a different IAID
     SCOPED_TRACE("CASE 2: Lease is known and belongs to this client, but to a different IAID");
 
@@ -708,6 +740,9 @@ Dhcpv6SrvTest::testReleaseReject(Lease::Type type, const IOAddress& addr) {
     l = LeaseMgrFactory::instance().getLease6(type, addr);
     ASSERT_TRUE(l);
 
+    // Verify we didn't decrement the stats counter
+    EXPECT_EQ(1, stat->getInteger().first);
+
     // CASE 3: Lease belongs to a client with different client-id
     SCOPED_TRACE("CASE 3: Lease belongs to a client with different client-id");
 
@@ -732,10 +767,50 @@ Dhcpv6SrvTest::testReleaseReject(Lease::Type type, const IOAddress& addr) {
     l = LeaseMgrFactory::instance().getLease6(type, addr);
     ASSERT_TRUE(l);
 
+    // Verify we didn't decrement the stats counter
+    EXPECT_EQ(1, stat->getInteger().first);
+
     // Finally, let's cleanup the database
     EXPECT_TRUE(LeaseMgrFactory::instance().deleteLease(addr));
 }
 
+void
+Dhcpv6SrvTest::testReceiveStats(uint8_t pkt_type, const std::string& stat_name) {
+
+    StatsMgr& mgr = StatsMgr::instance();
+    NakedDhcpv6Srv srv(0);
+
+    // Let's get a simple SOLICIT...
+    Pkt6Ptr pkt = PktCaptures::captureSimpleSolicit();
+
+    // And pretend it's packet of a different type
+    pkt->data_[0] = pkt_type;
+
+    // Check that those statistics are not set before the test
+    ObservationPtr pkt6_rcvd = mgr.getObservation("pkt6-received");
+    ObservationPtr tested_stat = mgr.getObservation(stat_name);
+    EXPECT_FALSE(pkt6_rcvd);
+    EXPECT_FALSE(tested_stat);
+
+    // Simulate that we have received that traffic
+    srv.fakeReceive(pkt);
+
+    // Server will now process to run its normal loop, but instead of calling
+    // IfaceMgr::receive6(), it will read all packets from the list set by
+    // fakeReceive()
+    srv.run();
+
+    // All expected statstics must be present.
+    pkt6_rcvd = mgr.getObservation("pkt6-received");
+    tested_stat = mgr.getObservation(stat_name);
+    ASSERT_TRUE(pkt6_rcvd);
+    ASSERT_TRUE(tested_stat);
+
+    // They also must have expected values.
+    EXPECT_EQ(1, pkt6_rcvd->getInteger().first);
+    EXPECT_EQ(1, tested_stat->getInteger().first);
+}
+
 void
 Dhcpv6SrvTest::configure(const std::string& config) {
     configure(config, srv_);
@@ -756,6 +831,48 @@ Dhcpv6SrvTest::configure(const std::string& config, NakedDhcpv6Srv& srv) {
     CfgMgr::instance().commit();
 }
 
+NakedDhcpv6SrvTest::NakedDhcpv6SrvTest()
+: rcode_(-1) {
+    // it's ok if that fails. There should not be such a file anyway
+    unlink(DUID_FILE);
+
+    const isc::dhcp::IfaceMgr::IfaceCollection& ifaces =
+        isc::dhcp::IfaceMgr::instance().getIfaces();
+
+    // There must be some interface detected
+    if (ifaces.empty()) {
+        // We can't use ASSERT in constructor
+        ADD_FAILURE() << "No interfaces detected.";
+    }
+
+    valid_iface_ = (*ifaces.begin())->getName();
+
+    // Let's wipe all existing statistics.
+    isc::stats::StatsMgr::instance().removeAll();
+}
+
+NakedDhcpv6SrvTest::~NakedDhcpv6SrvTest() {
+    // Let's wipe all existing statistics.
+    isc::stats::StatsMgr::instance().removeAll();
+
+    // Let's clean up if there is such a file.
+    unlink(DUID_FILE);
+    isc::hooks::HooksManager::preCalloutsLibraryHandle()
+        .deregisterAllCallouts("buffer6_receive");
+    isc::hooks::HooksManager::preCalloutsLibraryHandle()
+        .deregisterAllCallouts("buffer6_send");
+    isc::hooks::HooksManager::preCalloutsLibraryHandle()
+        .deregisterAllCallouts("lease6_renew");
+    isc::hooks::HooksManager::preCalloutsLibraryHandle()
+        .deregisterAllCallouts("lease6_release");
+    isc::hooks::HooksManager::preCalloutsLibraryHandle()
+        .deregisterAllCallouts("pkt6_receive");
+    isc::hooks::HooksManager::preCalloutsLibraryHandle()
+        .deregisterAllCallouts("pkt6_send");
+    isc::hooks::HooksManager::preCalloutsLibraryHandle()
+        .deregisterAllCallouts("subnet6_select");
+}
+
 // Generate IA_NA option with specified parameters
 boost::shared_ptr<Option6IA>
 NakedDhcpv6SrvTest::generateIA(uint16_t type, uint32_t iaid, uint32_t t1,
@@ -828,8 +945,8 @@ NakedDhcpv6SrvTest::checkIA_NAStatusCode(
     EXPECT_EQ(expected_t1, ia->getT1());
     EXPECT_EQ(expected_t2, ia->getT2());
 
-    isc::dhcp::OptionCustomPtr status =
-        boost::dynamic_pointer_cast<isc::dhcp::OptionCustom>
+    isc::dhcp::Option6StatusCodePtr status =
+        boost::dynamic_pointer_cast<isc::dhcp::Option6StatusCode>
         (ia->getOption(D6O_STATUS_CODE));
 
     // It is ok to not include status success as this is the default
@@ -846,9 +963,10 @@ NakedDhcpv6SrvTest::checkIA_NAStatusCode(
         // status code option content is just a text explanation
         // what went wrong.
         EXPECT_EQ(static_cast<uint16_t>(expected_status_code),
-                  status->readInteger<uint16_t>(0));
+                  status->getStatusCode());
     }
 }
 
+
 }; // end of isc::test namespace
 }; // end of isc namespace
