@@ -22,8 +22,10 @@
 #include <dhcp/option6_client_fqdn.h>
 #include <dhcp/option6_ia.h>
 #include <dhcp/option6_iaaddr.h>
+#include <dhcp/option6_status_code.h>
 #include <dhcp/option_int_array.h>
 #include <dhcpsrv/lease.h>
+#include <dhcp/tests/iface_mgr_test_config.h>
 
 #include <dhcp6/tests/dhcp6_test_utils.h>
 #include <boost/pointer_cast.hpp>
@@ -255,16 +257,9 @@ public:
     /// Status Code option.
     ///
     /// @return An object representing the Status Code option.
-    OptionCustomPtr createStatusCode(const uint16_t code,
+    Option6StatusCodePtr createStatusCode(const uint16_t code,
                                      const std::string& msg) {
-        OptionDefinition def("status-code", D6O_STATUS_CODE, "record");
-        def.addRecordField("uint16");
-        def.addRecordField("string");
-        OptionCustomPtr opt_status(new OptionCustom(def, Option::V6));
-        opt_status->writeInteger(code);
-        if (!msg.empty()) {
-            opt_status->writeString(msg, 1);
-        }
+        Option6StatusCodePtr opt_status(new Option6StatusCode(code, msg));
         return (opt_status);
     }
 
@@ -329,7 +324,8 @@ public:
         // Create three IAs, each having different address.
         addIA(1234, IOAddress("2001:db8:1::1"), answer);
 
-        ASSERT_NO_THROW(srv_->processClientFqdn(question, answer));
+        AllocEngine::ClientContext6 ctx;
+        ASSERT_NO_THROW(srv_->processClientFqdn(question, answer, ctx));
         Option6ClientFqdnPtr answ_fqdn = boost::dynamic_pointer_cast<
             Option6ClientFqdn>(answer->getOption(D6O_CLIENT_FQDN));
         ASSERT_TRUE(answ_fqdn);
@@ -393,6 +389,7 @@ public:
                             const std::string& exp_hostname,
                             const uint8_t client_flags =
                                 Option6ClientFqdn::FLAG_S,
+                            const IOAddress expected_address = IOAddress("2001:db8:1:1::dead:beef"),
                             const bool include_oro = true) {
         // Create a message of a specified type, add server id and
         // FQDN option.
@@ -440,7 +437,7 @@ public:
         ASSERT_TRUE(addr);
 
         // Check that we have got the address we requested.
-        checkIAAddr(addr, IOAddress("2001:db8:1:1::dead:beef"),
+        checkIAAddr(addr, expected_address,
                     Lease::TYPE_NA);
 
         if (msg_type != DHCPV6_SOLICIT) {
@@ -472,7 +469,7 @@ public:
     /// NameChangeRequest.
     /// @param dhcid An expected DHCID value.
     /// @note This value is the value that is produced by
-    /// dhcp_ddns::D2Dhcid::crateDigest() with the appropriate arguments. This
+    /// dhcp_ddns::D2Dhcid::createDigest() with the appropriate arguments. This
     /// method uses encryption tools to produce the value which cannot be
     /// easily duplicated by hand.  It is more or less necessary to generate
     /// these values programmatically and place them here. Should the
@@ -482,12 +479,15 @@ public:
     /// NameChangeRequest expires.
     /// @param len A valid lifetime of the lease associated with the
     /// NameChangeRequest.
+    /// @param fqdn The expected string value of the FQDN, if blank the
+    /// check is skipped
     void verifyNameChangeRequest(const isc::dhcp_ddns::NameChangeType type,
                                  const bool reverse, const bool forward,
                                  const std::string& addr,
                                  const std::string& dhcid,
                                  const uint16_t expires,
-                                 const uint16_t len) {
+                                 const uint16_t len,
+                                 const std::string& fqdn="") {
         NameChangeRequestPtr ncr;
         ASSERT_NO_THROW(ncr = d2_mgr_.peekAt(0));
         ASSERT_TRUE(ncr);
@@ -501,16 +501,46 @@ public:
         EXPECT_EQ(len, ncr->getLeaseLength());
         EXPECT_EQ(isc::dhcp_ddns::ST_NEW, ncr->getStatus());
 
+        if (! fqdn.empty()) {
+           EXPECT_EQ(fqdn, ncr->getFqdn());
+        }
+
+
         // Process the message off the queue
         ASSERT_NO_THROW(d2_mgr_.runReadyIO());
     }
 
+    /// @brief Updates inherited subnet and pool members
+    ///
+    /// Hack added to set subnet_ and pool_ members that are buried into lower
+    /// level tests such as checkLease(), so one can use "configure" functionality
+    /// rather than hand-building configured objects
+    /// @param subnet_idx Element index of the desired subnet
+    /// @param pool_idx Element index of the desired pool within the desired subnet
+    /// @param type lease type of desired pool
+    ///
+    void setSubnetAndPool(int subnet_idx, int pool_idx, Lease::Type type) {
+        ConstCfgSubnets6Ptr subnets = CfgMgr::instance().getCurrentCfg()->getCfgSubnets6();
+        ASSERT_TRUE(subnets);
+        const Subnet6Collection* subnet_col = subnets->getAll();
+        ASSERT_EQ(subnet_idx + 1, subnet_col->size());
+        subnet_ = subnet_col->at(subnet_idx);
+        ASSERT_TRUE(subnet_);
+
+        const PoolCollection& pool_col = subnet_->getPools(type);
+        ASSERT_EQ(pool_idx + 1, pool_col.size());
+        PoolPtr pool  = (subnet_->getPools(type)).at(pool_idx);
+        ASSERT_TRUE(pool);
+        pool_ = boost::dynamic_pointer_cast<Pool6>(pool);
+        ASSERT_TRUE(pool);
+    }
+
     // Holds a lease used by a test.
     Lease6Ptr lease_;
 
 };
 
-// A set of tests verifying server's behaviour when it receives the DHCPv6
+// A set of tests verifying server's behavior when it receives the DHCPv6
 // Client Fqdn Option.
 
 // Test server's response when client requests that server performs AAAA update.
@@ -695,7 +725,8 @@ TEST_F(FqdnDhcpv6SrvTest, createRemovalNameChangeRequestFwdRev) {
     // as if we typed domain-name in lower case.
     lease_->hostname_ = "MYHOST.example.com.";
 
-    ASSERT_NO_THROW(srv_->createRemovalNameChangeRequest(lease_));
+    Pkt6Ptr pkt(new Pkt6(DHCPREQUEST, 1234));
+    ASSERT_NO_THROW(srv_->createRemovalNameChangeRequest(pkt, lease_));
 
     ASSERT_EQ(1, d2_mgr_.getQueueSize());
     verifyNameChangeRequest(isc::dhcp_ddns::CHG_REMOVE, true, true,
@@ -720,7 +751,8 @@ TEST_F(FqdnDhcpv6SrvTest, noRemovalsWhenDisabled) {
     lease_->hostname_ = "MYHOST.example.com.";
 
     // When DDNS is disabled an attempt to send a request will throw.
-    ASSERT_NO_THROW(srv_->createRemovalNameChangeRequest(lease_));
+    Pkt6Ptr pkt(new Pkt6(DHCPREQUEST, 1234));
+    ASSERT_NO_THROW(srv_->createRemovalNameChangeRequest(pkt, lease_));
 }
 
 
@@ -731,7 +763,8 @@ TEST_F(FqdnDhcpv6SrvTest, createRemovalNameChangeRequestRev) {
     lease_->fqdn_rev_ = true;
     lease_->hostname_ = "myhost.example.com.";
 
-    ASSERT_NO_THROW(srv_->createRemovalNameChangeRequest(lease_));
+    Pkt6Ptr pkt(new Pkt6(DHCPREQUEST, 1234));
+    ASSERT_NO_THROW(srv_->createRemovalNameChangeRequest(pkt, lease_));
 
     ASSERT_EQ(1, d2_mgr_.getQueueSize());
 
@@ -749,7 +782,8 @@ TEST_F(FqdnDhcpv6SrvTest, createRemovalNameChangeRequestNoUpdate) {
     lease_->fqdn_fwd_ = false;
     lease_->fqdn_rev_ = false;
 
-    ASSERT_NO_THROW(srv_->createRemovalNameChangeRequest(lease_));
+    Pkt6Ptr pkt(new Pkt6(DHCPREQUEST, 1234));
+    ASSERT_NO_THROW(srv_->createRemovalNameChangeRequest(pkt, lease_));
 
     ASSERT_EQ(0, d2_mgr_.getQueueSize());
 
@@ -762,7 +796,8 @@ TEST_F(FqdnDhcpv6SrvTest, createRemovalNameChangeRequestNoHostname) {
     lease_->fqdn_rev_ = true;
     lease_->hostname_ = "";
 
-    ASSERT_NO_THROW(srv_->createRemovalNameChangeRequest(lease_));
+    Pkt6Ptr pkt(new Pkt6(DHCPREQUEST, 1234));
+    ASSERT_NO_THROW(srv_->createRemovalNameChangeRequest(pkt, lease_));
 
     ASSERT_EQ(0, d2_mgr_.getQueueSize());
 
@@ -776,7 +811,8 @@ TEST_F(FqdnDhcpv6SrvTest, createRemovalNameChangeRequestWrongHostname) {
     lease_->fqdn_rev_ = true;
     lease_->hostname_ = "myhost..example.com.";
 
-    ASSERT_NO_THROW(srv_->createRemovalNameChangeRequest(lease_));
+    Pkt6Ptr pkt(new Pkt6(DHCPREQUEST, 1234));
+    ASSERT_NO_THROW(srv_->createRemovalNameChangeRequest(pkt, lease_));
 
     ASSERT_EQ(0, d2_mgr_.getQueueSize());
 
@@ -943,7 +979,8 @@ TEST_F(FqdnDhcpv6SrvTest, processRequestWithoutFqdn) {
     // In this case, we expect that the FQDN option will not be included
     // in the server's response. The testProcessMessage will check that.
     testProcessMessage(DHCPV6_REQUEST, "myhost.example.com",
-                       "myhost.example.com.", Option6ClientFqdn::FLAG_S, false);
+                       "myhost.example.com.", Option6ClientFqdn::FLAG_S,
+                       IOAddress("2001:db8:1:1::dead:beef"), false);
     ASSERT_EQ(1, d2_mgr_.getQueueSize());
     verifyNameChangeRequest(isc::dhcp_ddns::CHG_ADD, true, true,
                             "2001:db8:1:1::dead:beef",
@@ -957,7 +994,8 @@ TEST_F(FqdnDhcpv6SrvTest, processRequestWithoutFqdn) {
 TEST_F(FqdnDhcpv6SrvTest, processRequestEmptyFqdn) {
     testProcessMessage(DHCPV6_REQUEST, "",
                        "myhost-2001-db8-1-1--dead-beef.example.com.",
-                       Option6ClientFqdn::FLAG_S, false);
+                       Option6ClientFqdn::FLAG_S,
+                       IOAddress("2001:db8:1:1::dead:beef"), false);
     ASSERT_EQ(1, d2_mgr_.getQueueSize());
     verifyNameChangeRequest(isc::dhcp_ddns::CHG_ADD, true, true,
                             "2001:db8:1:1::dead:beef",
@@ -1050,5 +1088,151 @@ TEST_F(FqdnDhcpv6SrvTest, processClientDelegation) {
                             0, 4000);
 }
 
+// Verify that the host reservation is found and used. Lease host name and
+// FQDN should be the reservation hostname suffixed by the qualifying suffix.
+TEST_F(FqdnDhcpv6SrvTest, hostnameReservationSuffix) {
+    isc::dhcp::test::IfaceMgrTestConfig test_config(true);
+
+    string config_str = "{ "
+        "\"interfaces-config\": {"
+        "  \"interfaces\": [ \"*\" ]"
+        "},"
+        "\"valid-lifetime\": 4000, "
+        "\"preferred-lifetime\": 3000,"
+        "\"rebind-timer\": 2000, "
+        "\"renew-timer\": 1000, "
+        "\"subnet6\": [ "
+        " { "
+        "    \"subnet\": \"2001:db8:1::/48\", "
+        "    \"pools\": [ { \"pool\": \"2001:db8:1:1::/64\" } ],"
+        "    \"interface\" : \"eth0\" , "
+        "   \"reservations\": ["
+        "    {"
+        "        \"duid\": \"" + duid_->toText() + "\","
+        "        \"ip-addresses\": [ \"2001:db8:1:1::babe\" ],"
+        "        \"hostname\": \"alice\""
+        "    }"
+        "    ]"
+        " } ],"
+        " \"dhcp-ddns\" : {"
+        "     \"enable-updates\" : true, "
+        "     \"qualifying-suffix\" : \"example.com\" }"
+        "}";
+
+    configure(config_str);
+
+    // Update subnet_ and pool_ members after config
+    setSubnetAndPool(0, 0, Lease::TYPE_NA);
+
+    ASSERT_NO_THROW(srv_->startD2());
+
+    ASSERT_TRUE(CfgMgr::instance().ddnsEnabled());
+
+    // Verify that the host reservation is found and lease name/FQDN are
+    // formed properly from the host name and qualifying suffix.
+    testProcessMessage(DHCPV6_REQUEST, "myhost.example.com",
+                       "alice.example.com.", 1, IOAddress("2001:db8:1:1::babe"));
+
+    // Verify that NameChangeRequest is correct.
+    verifyNameChangeRequest(isc::dhcp_ddns::CHG_ADD, true, true,
+                            "2001:db8:1:1::babe",
+                            "000201E2EB74FB53A5778E74AFD43870ECA5"
+                            "4150B1F52B0CFED434802DA1259D6D3CA4",
+                            0, 4000, "alice.example.com.");
+}
+
+// Verify that the host reservation is found and used, rather than dynamic
+// Address.  Lease host name and FQDN should be the reservation hostname
+// without a qualifying suffix.
+TEST_F(FqdnDhcpv6SrvTest, hostnameReservationNoSuffix) {
+    isc::dhcp::test::IfaceMgrTestConfig test_config(true);
+
+    string config_str = "{ "
+        "\"interfaces-config\": {"
+        "  \"interfaces\": [ \"*\" ]"
+        "},"
+        "\"valid-lifetime\": 4000, "
+        "\"preferred-lifetime\": 3000,"
+        "\"rebind-timer\": 2000, "
+        "\"renew-timer\": 1000, "
+        "\"subnet6\": [ "
+        " { "
+        "    \"subnet\": \"2001:db8:1::/48\", "
+        "    \"pools\": [ { \"pool\": \"2001:db8:1:1::/64\" } ],"
+        "    \"interface\" : \"eth0\" , "
+        "   \"reservations\": ["
+        "    {"
+        "        \"duid\": \"" + duid_->toText() + "\","
+        "        \"ip-addresses\": [ \"2001:db8:1:1::babe\" ],"
+        "        \"hostname\": \"alice.example.com\""
+        "    }"
+        "    ]"
+        " } ],"
+        " \"dhcp-ddns\" : {"
+        "     \"enable-updates\" : true, "
+        "     \"qualifying-suffix\" : \"\" }"
+        "}";
+
+    configure(config_str);
+    // Update subnet_ and pool_ members after config
+    setSubnetAndPool(0, 0, Lease::TYPE_NA);
+
+    ASSERT_NO_THROW(srv_->startD2());
+
+    ASSERT_TRUE(CfgMgr::instance().ddnsEnabled());
+
+    testProcessMessage(DHCPV6_REQUEST, "myhost.example.com",
+                       "alice.example.com.", 1,
+                       IOAddress("2001:db8:1:1::babe"));
+
+    // Verify that NameChangeRequest is correct.
+    verifyNameChangeRequest(isc::dhcp_ddns::CHG_ADD, true, true,
+                            "2001:db8:1:1::babe",
+                            "000201E2EB74FB53A5778E74AFD43870ECA5"
+                            "4150B1F52B0CFED434802DA1259D6D3CA4",
+                            0, 4000, "alice.example.com.");
+
+}
+
+TEST_F(FqdnDhcpv6SrvTest, hostnameReservationDdnsDisabled) {
+    isc::dhcp::test::IfaceMgrTestConfig test_config(true);
+
+    string config_str = "{ "
+        "\"interfaces-config\": {"
+        "  \"interfaces\": [ \"*\" ]"
+        "},"
+        "\"valid-lifetime\": 4000, "
+        "\"preferred-lifetime\": 3000,"
+        "\"rebind-timer\": 2000, "
+        "\"renew-timer\": 1000, "
+        "\"subnet6\": [ "
+        " { "
+        "    \"subnet\": \"2001:db8:1::/48\", "
+        "    \"pools\": [ { \"pool\": \"2001:db8:1:1::/64\" } ],"
+        "    \"interface\" : \"eth0\" , "
+        "    \"reservations\": ["
+        "    {"
+        "        \"duid\": \"" + duid_->toText() + "\","
+        "        \"ip-addresses\": [ \"2001:db8:1:1::babe\" ],"
+        "        \"hostname\": \"alice\""
+        "    }"
+        "    ]"
+        " } ],"
+        " \"dhcp-ddns\" : {"
+        "     \"enable-updates\" : false, "
+        "     \"qualifying-suffix\" : \"disabled.example.com\" }"
+        "}";
+
+    configure(config_str);
+
+    // Update subnet_ and pool_ members after config
+    setSubnetAndPool(0, 0, Lease::TYPE_NA);
+
+    ASSERT_FALSE(CfgMgr::instance().ddnsEnabled());
+
+    testProcessMessage(DHCPV6_REQUEST, "myhost.example.com",
+                       "alice.disabled.example.com.", 0,
+                       IOAddress("2001:db8:1:1::babe"));
+}
 
 }   // end of anonymous namespace
