@@ -159,15 +159,21 @@ struct SocketInfo {
 /// returned by the OS kernel when the socket is opened. Hence, it is
 /// convenient to allocate the buffer when the socket is being opened and
 /// utilze it throughout the lifetime of the socket.
-class Iface {
+///
+/// In order to avoid potentially expensive copies of the @c Iface objects
+/// holding pre-allocated buffers and multiple containers, this class is
+/// noncopyable.
+class Iface : public boost::noncopyable {
 public:
 
     /// Maximum MAC address length (Infiniband uses 20 bytes)
     static const unsigned int MAX_MAC_LEN = 20;
 
+    /// @brief Address type.
+    typedef util::OptionalValue<asiolink::IOAddress> Address;
+
     /// Type that defines list of addresses
-    typedef
-    std::list<util::OptionalValue<asiolink::IOAddress> > AddressCollection;
+    typedef std::list<Address> AddressCollection;
 
     /// @brief Type that holds a list of socket information.
     ///
@@ -500,6 +506,8 @@ private:
     std::vector<uint8_t> read_buffer_;
 };
 
+typedef boost::shared_ptr<Iface> IfacePtr;
+
 /// @brief This type describes the callback function invoked when error occurs
 /// in the IfaceMgr.
 ///
@@ -543,8 +551,8 @@ public:
     //      2 maps (ifindex-indexed and name-indexed) and
     //      also hide it (make it public make tests easier for now)
 
-    /// Type that holds a list of interfaces.
-    typedef std::list<Iface> IfaceCollection;
+    /// Type that holds a list of pointers to interfaces.
+    typedef std::list<IfacePtr> IfaceCollection;
 
     /// IfaceMgr is a singleton class. This method returns reference
     /// to its sole instance.
@@ -582,14 +590,14 @@ public:
     /// @return true if direct response is supported.
     bool isDirectResponseSupported() const;
 
-    /// @brief Returns interfac specified interface index
+    /// @brief Returns interface specified interface index
     ///
     /// @param ifindex index of searched interface
     ///
     /// @return interface with requested index (or NULL if no such
     ///         interface is present)
     ///
-    Iface* getIface(int ifindex);
+    IfacePtr getIface(int ifindex);
 
     /// @brief Returns interface with specified interface name
     ///
@@ -597,8 +605,7 @@ public:
     ///
     /// @return interface with requested name (or NULL if no such
     ///         interface is present)
-    ///
-    Iface* getIface(const std::string& ifname);
+    IfacePtr getIface(const std::string& ifname);
 
     /// @brief Returns container with all interfaces.
     ///
@@ -966,9 +973,11 @@ public:
     void addExternalSocket(int socketfd, SocketCallback callback);
 
     /// @brief Deletes external socket
-
     void deleteExternalSocket(int socketfd);
 
+    /// @brief Deletes all external sockets.
+    void deleteAllExternalSockets();
+
     /// @brief Set packet filter object to handle sending and receiving DHCPv4
     /// messages.
     ///
@@ -1031,7 +1040,7 @@ public:
     /// @param iface reference to Iface object.
     /// @note This function must be public because it has to be callable
     /// from unit tests.
-    void addInterface(const Iface& iface) {
+    void addInterface(const IfacePtr& iface) {
         ifaces_.push_back(iface);
     }
 
@@ -1132,23 +1141,6 @@ protected:
     /// Control-buffer, used in transmission and reception.
     boost::scoped_array<char> control_buf_;
 
-    /// @brief A wrapper for OS-specific operations before sending IPv4 packet
-    ///
-    /// @param m message header (will be later used for sendmsg() call)
-    /// @param control_buf buffer to be used during transmission
-    /// @param control_buf_len buffer length
-    /// @param pkt packet to be sent
-    void os_send4(struct msghdr& m, boost::scoped_array<char>& control_buf,
-                  size_t control_buf_len, const Pkt4Ptr& pkt);
-
-    /// @brief OS-specific operations during IPv4 packet reception
-    ///
-    /// @param m message header (was used during recvmsg() call)
-    /// @param pkt packet received (some fields will be set here)
-    ///
-    /// @return true if successful, false otherwise
-    bool os_receive4(struct msghdr& m, Pkt4Ptr& pkt);
-
 private:
     /// @brief Identifies local network address to be used to
     /// connect to remote address.
