@@ -17,6 +17,7 @@
 
 #include <asiolink/io_address.h>
 #include <dhcp/option.h>
+#include <dhcp/duid.h>
 #include <util/buffer.h>
 #include <dhcp/option.h>
 #include <dhcp/classify.h>
@@ -95,25 +96,35 @@ public:
     /// Method with throw exception if packet parsing fails.
     virtual void unpack();
 
-    /// @brief performs sanity check on a packet.
+    /// @brief Returns text representation of the primary packet identifiers
     ///
-    /// This is usually performed after unpack(). It checks if packet is sane:
-    /// required options are present, fields have sane content etc.
-    /// For example verifies that DHCP_MESSAGE_TYPE is present and have
-    /// reasonable value. This method is expected to grow significantly.
-    /// It makes sense to separate unpack() and check() for testing purposes.
+    /// This method is intended to be used to provide a consistent way to
+    /// identify packets within log statements.  It is an instance-level
+    /// wrapper around static makeLabel(). See this method for string
+    /// content.
     ///
-    /// @todo It is called from unpack() directly. It should be separated.
+    /// @return string with text representation
+    std::string getLabel() const;
+
+    /// @brief Returns text representation of the given packet identifiers
     ///
-    /// Method will throw exception if anomaly is found.
-    void check();
+    /// @param hwaddr - hardware address to include in the string, it may be
+    /// NULL.
+    /// @param client_id - client id to include in the string, it may be NULL.
+    /// to include in the string
+    /// @param transid - numeric transaction id to include in the string
+    ///
+    /// @return string with text representation
+    static std::string makeLabel(const HWAddrPtr& hwaddr,
+                                 const ClientIdPtr& client_id,
+                                 const uint32_t transid);
 
     /// @brief Returns text representation of the packet.
     ///
     /// This function is useful mainly for debugging.
     ///
     /// @return string with text representation
-    std::string toText();
+    std::string toText() const;
 
     /// @brief Returns the size of the required buffer to build the packet.
     ///
@@ -220,6 +231,22 @@ public:
     /// @param type message type to be set
     void setType(uint8_t type);
 
+    /// @brief Returns name of the DHCP message for a given type number.
+    ///
+    /// @param type DHCPv4 message type which name should be returned.
+    ///
+    /// @return Pointer to the "const" string containing DHCP message name.
+    /// If the message type is unsupported the "UNKNOWN" is returned.
+    /// The caller must not release the returned pointer.
+    static const char* getName(const uint8_t type);
+
+    /// @brief Returns name of the DHCP message.
+    ///
+    /// @return Pointer to the "const" string containing DHCP message name.
+    /// If the message type is unsupported the "UNKNOWN" is returned.
+    /// The caller must not release the returned pointer.
+    const char* getName() const;
+
     /// @brief Returns sname field
     ///
     /// Note: This is 64 bytes long field. It doesn't have to be
@@ -332,17 +359,11 @@ public:
     /// This function returns a boolean value which indicates whether a DHCPv4
     /// message has been relayed (if true is returned) or not (if false).
     ///
-    /// This function uses a combination of Giaddr and Hops. It is expected that
-    /// if Giaddr is not 0, the Hops is greater than 0. In this case the message
-    /// is considered relayed. If Giaddr is 0, the Hops value must also be 0. In
-    /// this case the message is considered non-relayed. For any other
-    /// combination of Giaddr and Hops, an exception is thrown to indicate that
-    /// the message is malformed.
+    /// The message is considered relayed if the giaddr field is non-zero and
+    /// non-broadcast.
     ///
     /// @return Boolean value which indicates whether the message is relayed
     /// (true) or non-relayed (false).
-    /// @throw isc::BadValue if invalid combination of Giaddr and Hops values is
-    /// found.
     bool isRelayed() const;
 
     /// @brief That's the data of input buffer used in RX packet.
